import numpy as np
from h5py import File
from typing import Tuple


class ImageRankHistogram:
    def __init__(self, hdfstore: File) -> None:
        self.imgnames = hdfstore["/feature_output/imgname"][:].astype("U")
        self.objnames = np.unique(np.array([c.split("-")[0] for c in self.imgnames]))

    def get_imagerank_histogram(
        self,
        hdfstore: File,
        ax: str,
        num_objs: int = 200,
        post_processed: bool = False,
        pp_exclusion: str = "soft",
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        if post_processed:
            key_head = "/pairwise_correlation/postprocessed/{}".format(pp_exclusion)
            key_head2 = "postprocessed_{}/{}".format(num_objs, pp_exclusion)
        else:
            key_head = "/pairwise_correlation/original"
            key_head2 = "original_{}".format(num_objs)

        obj_idxs = hdfstore[key_head + "_{}_obj_idx".format(num_objs)][:]
        objs = self.objnames[obj_idxs]

        all_imagerank_data = []
        for obj in objs:
            obj_ax_key = "/{}/{}/{}".format(key_head2, obj, ax)
            imagerank = hdfstore[obj_ax_key + "/sameobj_imgrank"][:]
            all_imagerank_data.append(imagerank)

        all_imagerank_data = np.concatenate(all_imagerank_data)

        stdev = np.nanstd(all_imagerank_data[all_imagerank_data != 0])
        avg = np.nanstd(all_imagerank_data[all_imagerank_data != 0])
        max_range = int(avg + 3 * stdev)
        # if max_range > 50:
        #     # bins = np.linspace(1, int(avg+3*stdev), 50)
        #     # all_imagerank_data[all_imagerank_data> int(avg+3*stdev)] = int(avg+3*stdev)

        # else:
        bins = np.linspace(1, 40, 40)
        all_imagerank_data[all_imagerank_data > 40] = 40

        bins = np.insert(bins, 0, 0)

        histcounts = np.apply_along_axis(
            lambda a: np.histogram(a, bins=bins)[0], 0, all_imagerank_data
        ).T

        def normalize_histogram(histcount, bins):
            dx = np.diff(bins)
            total_area = np.multiply(histcount, dx).sum()
            return histcount / total_area

        density_hist = np.apply_along_axis(
            lambda a: normalize_histogram(a, bins), 1, histcounts
        )
        return histcounts, density_hist, bins

    def get_objrank_histogram(
        self,
        hdfstore: File,
        ax: str,
        num_objs: int = 200,
        post_processed: bool = False,
        pp_exclusion: str = "soft",
    ) -> Tuple[np.ndarray, np.ndarray]:
        if post_processed:
            key_head = "/pairwise_correlation/postprocessed/{}".format(pp_exclusion)
            key_head2 = "postprocessed_{}/{}".format(num_objs, pp_exclusion)
        else:
            key_head = "/pairwise_correlation/original"
            key_head2 = "original_{}".format(num_objs)

        obj_idxs = hdfstore[key_head + "_{}_obj_idx".format(num_objs)][:]
        objs = self.objnames[obj_idxs]

        all_objrank_data = []
        for obj in objs:
            obj_ax_key = "/{}/{}/{}".format(key_head2, obj, ax)
            objrank = hdfstore[obj_ax_key + "/sameobj_objrank"][:]
            all_objrank_data.append(objrank)

        all_objrank_data = np.concatenate(all_objrank_data)
        bins = np.linspace(1, 40, 40)
        all_objrank_data[all_objrank_data > 40] = 40

        bins = np.insert(bins, 0, 0)

        histcounts = np.apply_along_axis(
            lambda a: np.histogram(a, bins=bins)[0], 0, all_objrank_data
        ).T

        def normalize_histogram(histcount, bins):
            dx = np.diff(bins)
            total_area = np.multiply(histcount, dx).sum()
            return histcount / total_area

        density_hist = np.apply_along_axis(
            lambda a: normalize_histogram(a, bins), 1, histcounts
        )
        return histcounts, density_hist, bins



import csv


def check_and_retrieve_stored_feature(imgname, imgname_list, feat_list):
    try:
        idx = imgname_list.index(imgname)
        feat = feat_list[idx]
    except ValueError:
        feat = "Not Found"

    return feat


def write_row_to_hdf(df, hdfstore, data_key):
    hdfstore.append(data_key, df, format="table")


def write_row_to_csv(row, csvfile):
    with open(csvfile, "a") as f:
        wrtr = csv.writer(f, delimiter=",", quotechar='"')
        wrtr.writerow(row)


def get_row_count(csvfile):
    with open(csvfile, "r") as f:
        reader = csv.reader(f)
        row_count = sum(1 for r in reader)

    return row_count



import torch.nn as nn


class GetModelIntermediateLayer(nn.Module):
    def __init__(self, original_model, layerindex):
        super(GetModelIntermediateLayer, self).__init__()
        self.features = nn.Sequential(*list(original_model.children())[:layerindex])

    def forward(self, x):
        x = self.features(x)
        return x



import torchvision.datasets as datasets
from torch.utils.data import Dataset
from itertools import combinations
import math
import psutil


class CombinationDataset(Dataset):
    def __init__(self, dataset):
        self.dataset = dataset
        self.comb = list(combinations(dataset, 2))

    def __getitem__(self, index):
        img1, img2 = self.comb[index]
        return img1, img2

    def __len__(self):
        return len(self.comb)

    def cut_dataset(self, index):
        self.comb = self.comb[index:]


class ImageFolderWithPaths(datasets.ImageFolder):
    """Custom dataset that includes image file paths. Extends
    torchvision.datasets.ImageFolder
    """

    # override the __getitem__ method. this is the method that dataloader calls
    def __getitem__(self, index):
        # this is what ImageFolder normally returns
        original_tuple = super(ImageFolderWithPaths, self).__getitem__(index)
        # the image file path
        path = self.imgs[index][0]
        # make a new tuple that includes original and the path
        tuple_with_path = original_tuple + (path,)
        return tuple_with_path


class FeatureTensorDatasetWithImgName(Dataset):
    def __init__(self, feature_tensor, img_name_array):
        self.feature_tensor = feature_tensor
        self.imgnames = img_name_array

    def __getitem__(self, index):
        feat = self.feature_tensor[index, :]
        imgname = self.imgnames[index]
        return imgname, feat

    def __len__(self):
        return len(self.imgnames)


class PermutationIndexDataset(Dataset):
    def __init__(self, datalen):
        self.datalen = datalen

    def __getitem__(self, index):
        idx1 = int(math.floor(index / self.datalen))
        idx2 = index % self.datalen
        return idx1, idx2


class OriginalandPostProcessedPairsDataset(Dataset):
    def __init__(self, original_feat_dataset, postprocessed_feat_dataset):
        self.original = original_feat_dataset
        self.postprocessed = postprocessed_feat_dataset
        self.datalen = len(self.postprocessed)

    def __getitem__(self, index):
        idx1 = int(math.floor(index / self.datalen))
        idx2 = index % self.datalen
        s1 = self.original[idx1]
        s2 = self.postprocessed[idx2]
        return (idx1, s1), (idx2, s2)

    def __len__(self):
        return len(self.original) ** 2


class PermutationPairsDataset(Dataset):
    def __init__(self, original_feat_dataset, postprocessed=None):
        self.original = original_feat_dataset
        self.datalen = len(self.original)
        self.postprocessed = postprocessed

    def __getitem__(self, index):
        idx1 = int(math.floor(index / self.datalen))
        idx2 = index % self.datalen
        s1 = self.original[idx1]
        if self.postprocessed is not None:
            s2 = self.postprocessed[idx2]
        else:
            s2 = self.original[idx2]
        return (idx1, s1), (idx2, s2)

    def __len__(self):
        return len(self.original) ** 2


class HDFDataset(Dataset):
    def __init__(self, hdfstore, mem_usage=0.85):
        self.hdfstore = hdfstore
        self.datalen = len(self.hdfstore)
        self.pull_data_to_cache(mem_usage)
        if not self.all_in_cache:
            print("initializing placeholder cache list")
            self.cache_length = int(
                psutil.virtual_memory().available * 0.85 / self.hdfstore[0].nbytes
            )
            self.in_cache_idx = [None] * self.cache_length
            self.in_cache = [None] * self.cache_length
            self.cache_counter = 0

    def __getitem__(self, index):
        if not self.all_in_cache:
            if index in self.in_cache_idx:
                return self.in_cache[self.in_cache_idx.index(index)]
            else:
                self.in_cache_idx[self.cache_counter] = index
                data = self.hdfstore[index]
                self.in_cache[self.cache_counter] = data
                self.cache_counter += 1
                self.cache_counter %= self.cache_length
                return data
        return self.hdfstore[index]

    def __len__(self):
        return self.datalen

    def pull_data_to_cache(self, mem_usage):
        single_row = self.hdfstore[0]
        if (
            psutil.virtual_memory().available * mem_usage
            < single_row.nbytes * self.datalen
        ):
            print("Not enough memory to pull data to cache")
            self.all_in_cache = False
        else:
            print("Pulling data to cache")
            self.hdfstore = self.hdfstore[:]
            self.all_in_cache = True
            print("Done pulling data to cache")



import torch
import torch.nn as nn


def pearsonr(x, y):
    mean_x = torch.mean(x)
    mean_y = torch.mean(y)
    xm = x.sub(mean_x)
    ym = y.sub(mean_y)
    r_num = xm.dot(ym)
    r_den = torch.norm(xm, 2) * torch.norm(ym, 2)
    r_val = r_num / r_den
    return r_val


def pearsonr_batch(x_batch, y_batch, prenormalized=False):
    if not prenormalized:
        mean_x_b = torch.mean(x_batch, 1, True)
        mean_y_b = torch.mean(y_batch, 1, True)
        xm_b = x_batch.sub(mean_x_b)
        ym_b = y_batch.sub(mean_y_b)
        r_num_b = (xm_b * ym_b).sum(1, keepdim=True)
        r_den_b = torch.norm(xm_b, dim=1, keepdim=True) * torch.norm(
            ym_b, dim=1, keepdim=True
        )
        r_vals = r_num_b / r_den_b
    else:
        r_vals = (x_batch * y_batch).sum(1, keepdim=True)
    return r_vals


def ln_batch(x_batch, y_batch, n=1):
    lndist = nn.PairwiseDistance(p=n, keepdim=True)
    d = lndist(x_batch, y_batch)
    return d



from dataclasses import dataclass
from typing import List, Optional

# Dataset Parameters
@dataclass
class DatasetConfig:
    project_dir: str
    data_dir: str
    cr: bool
    cr_data_dir: Optional[str]


# Network inference parameters
@dataclass
class NetworkParams:
    name: str
    input_size: int
    batch_size: int
    num_workers: int
    gpu_num: int


@dataclass
class DistanceParams:
    metric: str


@dataclass
class AnalysisParams:
    match_mode: str
    cr_mode: Optional[str]


@dataclass
class PipelineParams:
    step1_input: str
    step1_output: str
    step1_recompute: bool
    step2_input: str
    step3_input: str
    step3_output: str
    step4_input: str
    step4_output: str


@dataclass
class ShapeYConfig:
    data: DatasetConfig
    network: NetworkParams
    graph: AnalysisParams
    pipeline: PipelineParams
    distance: DistanceParams



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm


class NNClassificationErrorV2:
    @staticmethod
    def generate_top1_error_data(
        hdfstore,
        objnames,
        ax,
        key_head="/original",
        within_category_error=False,
        distance: str = "correlation",
    ):
        def gather_info_same_obj_cat(hdf_obj_ax, original_obj, objs_same_cat):
            same_objcat_cvals = []
            for other_obj in objs_same_cat:
                if original_obj != other_obj:
                    other_obj_cval = hdf_obj_ax[
                        "same_cat/{}/top1_cvals".format(other_obj)
                    ][:]
                    same_objcat_cvals.append(other_obj_cval)
                else:
                    top1_sameobj_cvals = hdf_obj_ax["top1_cvals"][:]
                    same_objcat_cvals.append(top1_sameobj_cvals)
            return np.array(same_objcat_cvals)

        # data holder
        top1_error_per_obj = []
        num_correct_allobj = []
        total_count = []

        for obj in objnames:
            key_obj = key_head + "/" + obj
            g = hdfstore[key_obj + "/" + ax]
            top1_excdist = g["top1_cvals"][
                :
            ]  # 1st dim = list of imgs in series, 2nd dim = exclusion dists, vals = top1 cvals with exc dist
            top1_other = g["top1_cvals_otherobj"][
                :
            ]  # 1st dim = list of imgs in series, vals = top1 cvals excluding the same obj

            # if within_category_error = True, you consider a match to another obj in the same obj category a correct answer
            if within_category_error:
                obj_cat = obj.split("_")[0]
                in_same_objcat = np.array(
                    [obj_cat == other_obj.split("_")[0] for other_obj in objnames]
                )

                same_objcat_cvals = gather_info_same_obj_cat(
                    g, obj, objnames[in_same_objcat]
                )  # 1st dim = different objs in same obj cat, 2nd dim = imgs, 3rd dim = exclusion dist in ax
                top_per_obj_cvals = g["top1_per_obj_cvals"][:]
                # zero out objs in same obj category
                same_obj_mask = np.tile(in_same_objcat[objnames != obj], (11, 1))
                if distance == "correlation":
                    # top 1 other object category
                    top_per_obj_cvals[same_obj_mask] = 0
                    top1_other_cat_cvals = np.max(top_per_obj_cvals, axis=1)
                    comparison_mask = np.tile(top1_other_cat_cvals, (11, 1)).T
                    # top 1 same obj category with exclusion
                    top1_same_cat_cvals = np.max(same_objcat_cvals, axis=0)
                    larger_than = np.greater(top1_same_cat_cvals, comparison_mask)
                else:
                    # top 1 other object category
                    top_per_obj_cvals[same_obj_mask] = np.nan
                    top1_other_cat_cvals = np.nanmin(top_per_obj_cvals, axis=1)
                    comparison_mask = np.tile(top1_other_cat_cvals, (11, 1)).T
                    # top 1 same obj category with exclusion
                    top1_same_cat_cvals = np.nanmin(same_objcat_cvals, axis=0)
                    larger_than = np.less(top1_same_cat_cvals, comparison_mask)
            else:
                comparison_mask = np.tile(top1_other, (11, 1)).T
                # compare if the largest cval for same obj is larger than the top1 cval for other objs
                if distance == "correlation":
                    larger_than = np.greater(top1_excdist, comparison_mask)
                else:
                    larger_than = np.less(top1_excdist, comparison_mask)

            correct = larger_than.sum(axis=0)
            total_sample = 11 - np.isnan(top1_excdist).sum(axis=0)
            top1_error = (total_sample - correct) / total_sample
            top1_error_per_obj.append(top1_error)
            num_correct_allobj.append(correct)
            total_count.append(total_sample)
        # compute average over all obj
        num_correct_allobj = np.array(num_correct_allobj).sum(axis=0)
        total_count = np.array(total_count).sum(axis=0)
        top1_error_mean = (total_count - num_correct_allobj) / total_count
        return top1_error_per_obj, top1_error_mean, num_correct_allobj, total_count

    @staticmethod
    def plot_top1_err_per_axis(
        top1_error_per_obj, top1_error_mean, legend_list, title, num_obj=200
    ):
        mStyles = [
            ".",
            ",",
            "v",
            "^",
            "<",
            ">",
            "1",
            "2",
            "3",
            "4",
            "8",
            "s",
            "p",
            "P",
            "*",
            "h",
            "H",
            "+",
            "x",
            "X",
        ]
        colors = cm.get_cmap("tab20", 20)
        figs, axes = plt.subplots(1, 1)
        all_data = np.array(top1_error_per_obj)
        axes.errorbar(
            np.linspace(0, 10, 11),
            top1_error_mean.T,
            yerr=[
                top1_error_mean.T - np.quantile(all_data, 0.25, axis=0),
                np.quantile(all_data, 0.75, axis=0) - top1_error_mean.T,
            ],
            linestyle="-",
            alpha=0.8,
            c="red",
            linewidth=2.5,
            marker="o",
            label="average",
            zorder=21,
            capsize=2,
        )

        obj_counter = -1
        objname = ""
        legend_reduced = list(set([l.split("_")[0] for l in legend_list]))
        offsets = np.linspace(-0.35, 0.35, 20)
        for idx, obj_cat in enumerate(legend_reduced):
            offset = offsets[idx]
            data = np.array(
                top1_error_per_obj[
                    int(num_obj / 20) * idx : int(num_obj / 20) * (idx + 1)
                ]
            )
            x = np.linspace(-1 + offset, 9 + offset, 11)
            tile_dim = data.shape[0]
            axes.scatter(
                np.tile(x, (tile_dim, 1)),
                data,
                marker=mStyles[idx],
                color=colors(idx),
                s=2,
                alpha=1,
            )
            axes.errorbar(
                x,
                data.mean(axis=0),
                yerr=[
                    data.mean(axis=0) - np.quantile(data, 0.25, axis=0),
                    np.quantile(data, 0.75, axis=0) - data.mean(axis=0),
                ],
                linestyle="--",
                alpha=0.9,
                linewidth=0.8,
                marker=mStyles[idx],
                color=colors(idx),
                label=obj_cat,
                capsize=2,
            )
        axes.set_ylim([-0.05, 1.0])
        axes.set_xlim([-0.5, 10.5])
        major_ticks = np.arange(0, 11, 1)
        minor_ticks = np.arange(-0.5, 10.5, 1)
        axes.set_xticks(major_ticks, minor=False)
        axes.set_xticks(minor_ticks, minor=True)
        axes.set_yticks([-0.02, 1.02], minor=True)
        axes.grid(linestyle="--", alpha=0.5, which="major")
        axes.tick_params("x", length=0, which="major")
        axes.tick_params("both", labelsize=14)
        axes.grid(linestyle="-", alpha=1, which="minor")
        for i in range(int(len(minor_ticks) / 2)):
            axes.axvspan(
                minor_ticks[2 * i + 1],
                minor_ticks[2 * (i + 1)],
                facecolor="b",
                alpha=0.05,
            )
        figs.suptitle("Top1 Error with Exclusion in {}".format(title))
        axes.set_xlabel("Exclusion Distance")
        axes.set_ylabel("Top1 Nearest Neighbor Classification Error")
        axes.legend(bbox_to_anchor=(1.05, 1), loc="upper left", ncol=1, fontsize=16)
        return figs, axes

    @staticmethod
    def plot_top1_err_avgd(num_correct_ax, total_count_ax, legend):
        mStyles = ["v", "8", "s", "p", "P", "*", "h", "H", "+", "x", "X"]
        figs, axes = plt.subplots(1, 1)
        top1_err_ax = [(y - x) / y for (x, y) in zip(num_correct_ax, total_count_ax)]
        x = np.linspace(-1, 9, 11)
        for i in range(len(top1_err_ax)):
            axes.plot(
                x,
                top1_err_ax[i],
                "-",
                alpha=1,
                linewidth=2,
                marker=mStyles[i],
                markersize=10,
            )

        # plot avg
        # total_correct = np.array(num_correct_ax).sum(axis=0)
        # total_count = np.array(total_count_ax).sum(axis=0)
        # top1_avg = (total_count-total_correct)/total_count
        legend = list(legend)
        # legend.append('mean')
        # axes.plot(top1_avg, '-', alpha=0.8, c='red', linewidth=2.5, marker='o')
        axes.set_ylim([-0.05, 1.05])
        axes.set_xlim([-1.5, 9.5])
        axes.set_xticks(list(range(-1, 10)))
        axes.set_yticks([0.0, 0.2, 0.4, 0.6, 0.8, 1.0])
        axes.tick_params(axis="both", labelsize=20)
        axes.grid(linestyle="--", alpha=0.5)
        axes.legend(legend, loc="upper left", bbox_to_anchor=(-0.8, 0.97), fontsize=20)

        # change axes tick label
        figs.canvas.draw()
        labels = [item.get_text() for item in axes.get_xticklabels()]
        labels[0] = ""
        axes.set_xticklabels(labels)
        return figs, axes



from PIL import Image
import sys
import os
import numpy as np
import h5py
from itertools import combinations
import functools
from typing import Tuple
import re
import matplotlib
from tqdm import tqdm

matplotlib.use("Agg")

import matplotlib.pyplot as plt
from matplotlib import cm
from mpl_toolkits.axes_grid1 import ImageGrid as MplImageGrid
from os import path

PROJECT_DIR = path.dirname(__file__)
BLANK_IMG = path.join(PROJECT_DIR, "blank.png")


def blockPrint():
    sys.stdout = open(os.devnull, "w")


def enablePrint():
    sys.stdout = sys.__stdout__


class DataPrepImagePanels:
    def __init__(
        self,
        imgdir,
        datadir,
        num_series=31,
        num_img_single_series=11,
        num_obj_per_cat=10,
    ):
        (
            self.imgnames,
            self.objnames,
            self.obj_categories,
        ) = self.get_image_and_obj_names(datadir)
        self.datadir = datadir
        self.imgdir = imgdir
        self.num_series = num_series
        self.num_img_single_series = num_img_single_series
        self.num_obj_per_cat = num_obj_per_cat
        self.axis_of_interest = self.make_axis_of_interest()

    def make_axis_of_interest(self) -> list:
        axes = ["x", "y", "p", "r", "w"]
        axis_of_interest = []
        for choose in range(1, 7):
            for comb in combinations(axes, choose):
                axis_of_interest.append(functools.reduce(lambda a, b: a + b, comb))
        axis_of_interest.sort()
        return axis_of_interest

    def get_image_and_obj_names(self, datadir):
        with h5py.File(datadir, "r") as hdfstore:
            imgnames = hdfstore["/feature_output/imgname"][:].astype("U")
            objnames = np.unique(np.array([c.split("-")[0] for c in imgnames]))
            obj_categories = imgnames
            obj_categories = [s.split("_")[0] for s in obj_categories]
            obj_categories = np.unique(np.array(obj_categories))
        return imgnames, objnames, obj_categories

    def map_to_imgname(self, idx):
        blank = np.where(idx == -1)
        imgs = self.imgnames[idx]
        imgs[blank] = "blank.png"
        return imgs

    def get_list_of_errors(
        self, obj: str, ax: str, exc_dist: int, key_head: str = "original"
    ) -> list:
        # returns list with errors [[img1: refimg, img2: best matching other obj]]
        with h5py.File(self.datadir, "r") as hdfstore:
            obj_ax_key = "/{}/{}/{}".format(key_head, obj, ax)
            top1_cvals_sameobj = hdfstore[obj_ax_key + "/top1_cvals"][:, exc_dist]
            top1_cvals_otherobj = hdfstore[obj_ax_key + "/top1_cvals_otherobj"][:]
            top1_idx_otherobj = hdfstore[obj_ax_key + "/top1_idx_otherobj"][:]
        incorrect = top1_cvals_sameobj < top1_cvals_otherobj
        incorrect[np.isnan(top1_cvals_sameobj)] = False
        c1, c2 = self.get_coord_corrmat(obj, ax)
        imgidxs = np.array(range(c1, c2))
        error_refimg = self.imgnames[imgidxs[incorrect]]
        error_cval_refimg = top1_cvals_sameobj[incorrect]
        error_matchimg = self.imgnames[top1_idx_otherobj[incorrect]]
        error_cval_matchimg = top1_cvals_otherobj[incorrect]
        # randomly select sample
        if len(error_refimg) > 3:
            idx = np.random.choice(len(error_refimg), 3, replace=False)
            idx.sort()
            error_refimg = error_refimg[idx]
            error_cval_refimg = error_cval_refimg[idx]
            error_matchimg = error_matchimg[idx]
            error_cval_matchimg = error_cval_matchimg[idx]

        # get all candidate images
        candidate_list = []
        candidate_cval_list = []
        for img in error_refimg:
            series_name = img.split("-")[1].split(".")[0]
            series_num = int(re.findall("\d+", series_name)[0]) - 1
            candidates, candidate_cvals, dists = self.get_all_candidate_images_sameobj(
                obj, ax, series_num, key_head=key_head
            )
            include = dists >= exc_dist
            candidates = candidates[include]
            candidate_cvals = candidate_cvals[include]
            if len(candidates) > 10:
                dists = dists[include]
                idx_order_dist = np.argsort(dists)[:10]
                candidates = candidates[idx_order_dist]
                candidate_cvals = candidate_cvals[idx_order_dist]
            candidate_list.append(candidates)
            candidate_cval_list.append(candidate_cvals)

        list_of_errors = list(
            map(list, zip(error_refimg, error_matchimg, candidate_list))
        )
        error_cvals = list(
            map(list, zip(error_cval_refimg, error_cval_matchimg, candidate_cval_list))
        )
        return list_of_errors, error_cvals

    def get_whole_error_list(
        self, ax: str, exc_dist: int, key_head: str = "original"
    ) -> Tuple[list, list]:
        first = True
        for obj in tqdm(self.objnames):
            if first:
                list_of_errors, error_cvals = self.get_list_of_errors(
                    obj, ax, exc_dist, key_head=key_head
                )
                first = False
            else:
                new_list_of_errors, new_error_cvals = self.get_list_of_errors(
                    obj, ax, exc_dist, key_head=key_head
                )
                list_of_errors += new_list_of_errors
                error_cvals += new_error_cvals
        return list_of_errors, error_cvals

    def get_coltf_sameobj_candidates(
        self, obj_name: str, axis: str
    ) -> Tuple[np.ndarray, np.ndarray]:
        same_obj_tf = np.char.startswith(self.imgnames, obj_name)
        series_names = np.array(
            [img.split("-")[1].split(".")[0] for img in self.imgnames]
        )
        same_series_tf = np.array([axis in s for s in series_names])
        col_tf = same_obj_tf & same_series_tf
        dists = np.array([int(re.findall(r"\d+", s)[0]) for s in series_names])
        dists = dists[col_tf]
        return col_tf, dists

    def get_all_candidate_images_sameobj(
        self, obj_name: str, axis: str, series_idx: int, key_head: str = "original"
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        cval_mat = self.grab_whole_rows(obj_name, axis, key_head=key_head)
        cval_row = cval_mat[series_idx, :]  # series index starts from 0!
        col_tf, dists = self.get_coltf_sameobj_candidates(obj_name, axis)
        dists = abs(dists - (series_idx + 1))
        candidates = self.imgnames[col_tf]
        candidate_cvals = cval_row[col_tf]
        return candidates, candidate_cvals, dists

    def get_top_matches_sameobj_with_excdist(
        self,
        candidates: np.ndarray,
        candidate_cvals: np.ndarray,
        dists: np.ndarray,
        excdist: int,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        candidates = candidates[dists >= excdist]
        candidate_cvals = candidate_cvals[dists >= excdist]
        dists = dists[dists >= excdist]
        sorted_idxs = np.argsort(-1 * candidate_cvals)
        candidates = np.take_along_axis(candidates, sorted_idxs, axis=0)
        candidate_cvals = np.take_along_axis(candidate_cvals, sorted_idxs, axis=0)
        dists = np.take_along_axis(dists, sorted_idxs, axis=0)
        return candidates, candidate_cvals, dists

    def get_coord_corrmat(self, obj_name, ax="all"):
        idx = np.where(self.objnames == obj_name)[0][0]
        if ax == "all":
            return (
                idx * self.num_img_single_series * self.num_series,
                (idx + 1) * self.num_img_single_series * self.num_series,
            )
        else:
            ax_idx = self.axis_of_interest.index(ax)
            return (
                idx * self.num_img_single_series * self.num_series
                + ax_idx * self.num_img_single_series,
                idx * self.num_img_single_series * self.num_series
                + (ax_idx + 1) * self.num_img_single_series,
            )

    def grab_whole_rows(self, objname, axis, key_head="original"):
        obj_ax_key = "/" + key_head + "/" + str(objname) + "/" + axis
        with h5py.File(self.datadir, "r") as hdfstore:
            r1, r2 = self.get_coord_corrmat(objname, ax=axis)
            whole_results = hdfstore["/pairwise_correlation/original"][r1:r2, :]
        return whole_results

    def map_top_per_obj_idx_to_global_idx(self, objname, top_per_obj_idx):
        top_per_obj_idx_mapped = np.copy(top_per_obj_idx)
        obj_idx = np.tile(
            np.linspace(
                0,
                top_per_obj_idx_mapped.shape[1] - 1,
                num=top_per_obj_idx_mapped.shape[1],
            ),
            (top_per_obj_idx_mapped.shape[0], 1),
        )
        obj_idx[:, np.where(self.objnames == objname)[0][0] :] = (
            obj_idx[:, np.where(self.objnames == objname)[0][0] :] + 1
        )
        obj_idx = obj_idx * self.num_img_single_series * self.num_series
        top_per_obj_idx_mapped = top_per_obj_idx_mapped + obj_idx
        top_per_obj_idx_mapped = top_per_obj_idx_mapped.astype(np.int)
        return top_per_obj_idx_mapped

    def get_best_matching_same_obj_with_exclusion(
        self, objname, axis, sample_row_idx, key_head="original"
    ):
        # sampled image name
        sampled_imgname = "{}-{}{:02d}.png".format(objname, axis, sample_row_idx + 1)
        with h5py.File(self.datadir, "r") as hdfstore:
            obj_ax_key = "/" + key_head + "/" + str(objname) + "/" + axis
            print("pulling {}".format(obj_ax_key))
            top1cvals_excdist = hdfstore[obj_ax_key + "/top1_cvals"][:]
            top1cvals_idx = hdfstore[obj_ax_key + "/top1_idx"][:]
        # row1 best matching same obj with exclusion distances
        invalid = np.where(top1cvals_idx == -1)
        top1cvals_idx = (
            top1cvals_idx
            + np.where(self.objnames == objname)[0][0]
            * self.num_img_single_series
            * self.num_series
        )
        top1cvals_idx[invalid] = -1
        row1_excdist_cvals = top1cvals_excdist[sample_row_idx, :]
        row1_imgnames = np.array(list(map(self.map_to_imgname, top1cvals_idx)))[
            sample_row_idx, :
        ]
        return sampled_imgname, (row1_excdist_cvals, row1_imgnames)

    def get_top11_best_matching_any_other_obj(
        self, objname, axis, sample_row_idx, key_head="original"
    ):
        # sampled image name
        sampled_imgname = "{}-{}{:02d}.png".format(objname, axis, sample_row_idx + 1)
        with h5py.File(self.datadir, "r") as hdfstore:
            obj_ax_key = "/" + key_head + "/" + str(objname) + "/" + axis
            print("pulling {}".format(obj_ax_key))
            top1_cvals_otherobj = hdfstore[obj_ax_key + "/top1_cvals_otherobj"][:]
            top1_idx_otherobj = hdfstore[obj_ax_key + "/top1_idx_otherobj"][:]
        # row2 top 11 best matching any obj
        whole_results = self.grab_whole_rows(objname, axis)
        c1, c2 = self.get_coord_corrmat(objname)
        whole_results[:, c1:c2] = 0
        top11_allobj_idx = np.argsort(-whole_results)[:, :11]
        top11_allobj = np.take_along_axis(whole_results, top11_allobj_idx, axis=1)
        top11_imgnames = np.array(list(map(self.map_to_imgname, top11_allobj_idx)))

        row2_top11_any_cvals = top11_allobj[sample_row_idx, :]
        row2_imgnames = top11_imgnames[sample_row_idx, :]
        return sampled_imgname, (row2_top11_any_cvals, row2_imgnames)

    def get_top11_best_matching_per_any_other_obj(
        self, objname, axis, sample_row_idx, key_head="original"
    ):
        # sampled image name
        sampled_imgname = "{}-{}{:02d}.png".format(objname, axis, sample_row_idx + 1)
        with h5py.File(self.datadir, "r") as hdfstore:
            obj_ax_key = "/" + key_head + "/" + str(objname) + "/" + axis
            print("pulling {}".format(obj_ax_key))
            top_per_obj_cvals = hdfstore[obj_ax_key + "/top1_per_obj_cvals"][:]
            top_per_obj_idx = hdfstore[obj_ax_key + "/top1_per_obj_idxs"][
                :
            ]  # top per obj index are indices within the object, thus needs to be mapped to a global index
        # row3 top 11 best matching per object
        row3_top_per_obj_idx = np.argsort(-top_per_obj_cvals)[:, :11]

        top_per_obj_idx_mapped = self.map_top_per_obj_idx_to_global_idx(
            objname, top_per_obj_idx
        )
        row3_img_idx = np.take_along_axis(
            top_per_obj_idx_mapped, row3_top_per_obj_idx, axis=1
        )

        row3_top_per_obj_cvals = np.take_along_axis(
            top_per_obj_cvals, row3_top_per_obj_idx, axis=1
        )[sample_row_idx, :]
        row3_imgnames = np.array(list(map(self.map_to_imgname, row3_img_idx)))[
            sample_row_idx, :
        ]
        return sampled_imgname, (row3_top_per_obj_cvals, row3_imgnames)

    def get_top11_best_matching_per_any_other_obj_category(
        self, objname, axis, sample_row_idx, key_head="original", keep_sameobj_cat=False
    ):
        # sampled image name
        sampled_imgname = "{}-{}{:02d}.png".format(objname, axis, sample_row_idx + 1)
        with h5py.File(self.datadir, "r") as hdfstore:
            obj_ax_key = "/" + key_head + "/" + str(objname) + "/" + axis
            print("pulling {}".format(obj_ax_key))
            top_per_obj_cvals = hdfstore[obj_ax_key + "/top1_per_obj_cvals"][:]
            top_per_obj_idx = hdfstore[obj_ax_key + "/top1_per_obj_idxs"][:]
        top_per_obj_idx_mapped = self.map_top_per_obj_idx_to_global_idx(
            objname, top_per_obj_idx
        )
        # row4 top 11 best matching per object category
        current_obj_cat = objname.split("_")[0]

        top_per_objcat_cvals = []
        top_per_objcat_idxs = []
        start_idx = 0
        for objcat in self.obj_categories:
            if objcat == current_obj_cat:
                end_idx = start_idx + self.num_obj_per_cat - 1
                if keep_sameobj_cat:
                    top_per_objcat_cvals.append(
                        top_per_obj_cvals[:, start_idx:end_idx].max(axis=1)
                    )
                    idxs = top_per_obj_cvals[:, start_idx:end_idx].argmax(axis=1)
                    top_per_objcat_idxs.append(
                        np.take_along_axis(
                            top_per_obj_idx_mapped[:, start_idx:end_idx],
                            np.expand_dims(idxs, axis=1),
                            axis=1,
                        )
                    )
            else:
                end_idx = start_idx + self.num_obj_per_cat
                top_per_objcat_cvals.append(
                    top_per_obj_cvals[:, start_idx:end_idx].max(axis=1)
                )
                idxs = top_per_obj_cvals[:, start_idx:end_idx].argmax(axis=1)
                top_per_objcat_idxs.append(
                    np.take_along_axis(
                        top_per_obj_idx_mapped[:, start_idx:end_idx],
                        np.expand_dims(idxs, axis=1),
                        axis=1,
                    )
                )
            start_idx = end_idx

        top_per_objcat_cvals = np.array(top_per_objcat_cvals, dtype=float).T
        top_per_objcat_idxs = np.squeeze(
            np.array(top_per_objcat_idxs, dtype=np.int64).T, axis=0
        )

        row4_top_per_objcat_idx = np.argsort(-top_per_objcat_cvals)[:, :11]
        row4_img_idx = np.take_along_axis(
            top_per_objcat_idxs, row4_top_per_objcat_idx, axis=1
        )

        row4_top_per_objcat_cvals = np.take_along_axis(
            top_per_objcat_cvals, row4_top_per_objcat_idx, axis=1
        )[sample_row_idx, :]
        row4_imgnames = np.array(list(map(self.map_to_imgname, row4_img_idx)))[
            sample_row_idx, :
        ]
        return sampled_imgname, (row4_top_per_objcat_cvals, row4_imgnames)

    def get_top_per_obj_row_with_exclusion(
        self, row_info_same_obj, row_info_top_per_obj, exclusion_dist
    ):
        # grab the top matching same obj with exclusion
        best_matching_same_obj = row_info_same_obj[1][exclusion_dist]
        best_matching_same_obj_cval = row_info_same_obj[0][exclusion_dist]
        # append and sort
        top_per_obj_cvals = np.append(
            row_info_top_per_obj[0], best_matching_same_obj_cval
        )
        top_per_obj_imgnames = np.append(
            row_info_top_per_obj[1], best_matching_same_obj
        )
        sorted = np.argsort(top_per_obj_cvals)
        top_per_obj_cvals = np.take_along_axis(top_per_obj_cvals, sorted[::-1], axis=0)
        top_per_obj_imgnames = np.take_along_axis(
            top_per_obj_imgnames, sorted[::-1], axis=0
        )
        # check for invalid sample with exclusion
        if np.isnan(top_per_obj_cvals[0]):
            top_per_obj_imgnames[:] = "blank.png"
        return (top_per_obj_cvals, top_per_obj_imgnames)

    def get_top_per_objcat_with_exclusion(
        self, row_info_top_per_obj, row_info_top_per_objcat, obj_cat
    ):
        same_obj_cat_tf = [obj_cat in s.split("_") for s in row_info_top_per_obj[1]]
        same_obj_cat_cval = 0.0
        for i, tf in enumerate(same_obj_cat_tf):
            if tf:
                if row_info_top_per_obj[0][i] > same_obj_cat_cval:
                    same_obj_cat_cval = row_info_top_per_obj[0][i]
                    same_obj_cat_imname = row_info_top_per_obj[1][i]
        if "same_obj_cat_imname" in locals():
            top_per_objcat_cvals = np.append(
                row_info_top_per_objcat[0], same_obj_cat_cval
            )
            top_per_objcat_imgnames = np.append(
                row_info_top_per_objcat[1], same_obj_cat_imname
            )
        else:
            top_per_objcat_cvals = row_info_top_per_objcat[0]
            top_per_objcat_imgnames = row_info_top_per_objcat[1]

        # remove duplicates if there is one
        same_obj_cat_tf = np.array(
            [obj_cat in s.split("_") for s in top_per_objcat_imgnames]
        )
        if same_obj_cat_tf.sum() > 1:
            same_obj_idx = np.where(same_obj_cat_tf == True)
            min_cval = 1.0
            remove_idx = 0
            for idx in same_obj_idx[0]:
                if top_per_objcat_cvals[idx] < min_cval:
                    min_cval = top_per_objcat_cvals[idx]
                    remove_idx = idx
            top_per_objcat_cvals = np.delete(top_per_objcat_cvals, remove_idx)
            top_per_objcat_imgnames = np.delete(top_per_objcat_imgnames, remove_idx)

        sorted = np.argsort(top_per_objcat_cvals)
        top_per_objcat_cvals = np.take_along_axis(
            top_per_objcat_cvals, sorted[::-1], axis=0
        )
        top_per_objcat_imgnames = np.take_along_axis(
            top_per_objcat_imgnames, sorted[::-1], axis=0
        )
        return (top_per_objcat_cvals, top_per_objcat_imgnames)

    def get_top_per_obj_whole_panel_with_exclusion(self, objname, axis, exclusion_dist):
        # grab all inputs
        blockPrint()
        row1_infos = [
            self.get_best_matching_same_obj_with_exclusion(objname, axis, samp_idx)
            for samp_idx in range(0, 11)
        ]
        row3_infos = [
            self.get_top11_best_matching_per_any_other_obj(objname, axis, samp_idx)
            for samp_idx in range(0, 11)
        ]
        enablePrint()
        # organize the infos
        top_per_obj_row_infos = [
            self.get_top_per_obj_row_with_exclusion(e[0][1], e[1][1], exclusion_dist)
            for e in zip(row1_infos, row3_infos)
        ]
        sampled_imgs = [e[0] for e in row1_infos]
        return sampled_imgs, top_per_obj_row_infos

    def get_top_per_objcat_whole_panel_with_exclusion(
        self, objname, axis, exclusion_dist
    ):
        # grab all inputs
        blockPrint()
        row1_infos = [
            self.get_best_matching_same_obj_with_exclusion(objname, axis, samp_idx)
            for samp_idx in range(0, 11)
        ]
        row3_infos = [
            self.get_top11_best_matching_per_any_other_obj(objname, axis, samp_idx)
            for samp_idx in range(0, 11)
        ]
        row4_infos = [
            self.get_top11_best_matching_per_any_other_obj_category(
                objname, axis, samp_idx, keep_sameobj_cat=True
            )
            for samp_idx in range(0, 11)
        ]
        enablePrint()
        # organize the infos
        # replace the best matching per objcat to self if larger than other category
        for i, (samp_img, (cval, imname)) in enumerate(row1_infos):
            if ~np.isnan(cval[exclusion_dist]):
                row3_infos[i] = (
                    samp_img,
                    (
                        np.insert(row3_infos[i][1][0], 0, cval[exclusion_dist]),
                        np.insert(row3_infos[i][1][1], 0, imname[exclusion_dist]),
                    ),
                )
        top_per_obj_row_infos = [
            self.get_top_per_objcat_with_exclusion(
                e[0][1], e[1][1], obj_cat=objname.split("_")[0]
            )
            for e in zip(row3_infos, row4_infos)
        ]
        sampled_imgs = [e[0] for e in row3_infos]
        # make nan for invalid samples
        for i, (_, (cval, _)) in enumerate(row1_infos):
            if np.isnan(cval[exclusion_dist]):
                top_per_obj_row_infos[i][1][:] = "blank.png"
        return sampled_imgs, top_per_obj_row_infos

    def get_whole_panel_single_sample(self, objname, axis, sample_idx):
        blockPrint()
        sampled_img, row1_info = self.get_best_matching_same_obj_with_exclusion(
            objname, axis, sample_idx
        )
        _, row2_info = self.get_top11_best_matching_any_other_obj(
            objname, axis, sample_idx
        )
        _, row3_info = self.get_top11_best_matching_per_any_other_obj(
            objname, axis, sample_idx
        )
        _, row4_info = self.get_top11_best_matching_per_any_other_obj_category(
            objname, axis, sample_idx
        )
        all_row_infos = [row1_info, row2_info, row3_info, row4_info]
        enablePrint()
        return sampled_img, all_row_infos

    def get_corr_fall_off_graph_infos_whole_panel(self, objname, axis, objcat=False):
        objcat = objname.split("_")[0]
        issameobjcat = np.array([objcat in obj for obj in self.objnames])
        objnames = self.objnames[issameobjcat]
        corr_fall_off_row_infos = []
        for obj in objnames:
            row = []
            for exc_dist in range(0, 11):
                if objcat:
                    (
                        sampled_imgs,
                        top_per_obj_row_infos,
                    ) = self.get_top_per_objcat_whole_panel_with_exclusion(
                        obj, axis, exc_dist
                    )
                else:
                    (
                        sampled_imgs,
                        top_per_obj_row_infos,
                    ) = self.get_top_per_obj_whole_panel_with_exclusion(
                        obj, axis, exc_dist
                    )
                row.append((sampled_imgs, top_per_obj_row_infos))
            corr_fall_off_row_infos.append(row)
        return corr_fall_off_row_infos


class ImageGridV2:
    def __init__(self, row: int, col: int, axis_pad: float = 0.05):
        self.row = row
        self.col = col
        self.axis_pad = axis_pad
        self.figure = plt.figure(figsize=(1.65 * self.col, 1.65 * self.row))
        self.grid = MplImageGrid(
            self.figure, 111, nrows_ncols=(self.row, self.col), axes_pad=self.axis_pad
        )

    def reset_figure(self):
        self.figure = plt.figure(figsize=(1.6 * self.col, 1.6 * self.row))
        self.grid = MplImageGrid(
            self.figure, 111, nrows_ncols=(self.row, self.col), axes_pad=self.axis_pad
        )

    def fill_panel(self, imnames, ylabels, annotations):
        for i, ax in enumerate(self.grid):
            ax.set_xticks([])
            ax.set_yticks([])
            r = i // self.col
            c = i % self.col
            im = imnames[r][c]
            im1 = Image.open(im)
            ax.imshow(im1)
            if c == 0:
                ylabel_separated = ylabels[r].split("-")
                if len(ylabel_separated) > 1:
                    ax.set_ylabel(
                        "{}\n{}".format(ylabel_separated[0], ylabel_separated[1]),
                        fontsize=7,
                        fontweight="bold",
                    )
                else:
                    ax.set_ylabel(ylabels[r], fontsize=20, fontweight="bold")
            t = annotations[r][c]
            ann_seperate = t.split(":")
            top_line_text = ann_seperate[0].split("-")
            if len(top_line_text) > 1:
                t = "{} \n {}".format(top_line_text[0], top_line_text[1])
            else:
                t = top_line_text[0]
            ax.text(
                128,
                5,
                t,
                color="yellow",
                fontsize=7,
                horizontalalignment="center",
                verticalalignment="top",
            )
            if len(ann_seperate) > 1:
                ax.text(
                    128,
                    250,
                    ann_seperate[1],
                    color="yellow",
                    fontsize=10,
                    horizontalalignment="center",
                )


class ImagePanelErrorDisplay(ImageGridV2):
    def __init__(self, datadir):
        super().__init__(10, 12)
        self.datadir = datadir

    def shorten_imgname(self, imgname):
        obj_codes = imgname.split("_")
        series_name = obj_codes[1].split("-")[1]
        series_name = series_name.split(".")[0]
        shortened_imgname = (
            obj_codes[0] + "_" + obj_codes[1][0:4] + "-{}".format(series_name)
        )
        return shortened_imgname

    def fill_error_panel(
        self,
        list_of_errors: list,
        cval_list: list,
        ax: str,
        exc_dist: int,
        annotate: bool = True,
    ) -> None:
        blank_img = BLANK_IMG
        imnames = []
        ylabels = []
        annotations = []
        for i, row in enumerate(list_of_errors):
            r = [self.datadir + row[0], self.datadir + row[1]]
            shortened_imgname = self.shorten_imgname(row[0])
            ylabels.append(shortened_imgname)
            ann = []
            ann.append("")
            if annotate:
                imname_short = self.shorten_imgname(row[1])
                ann.append("{}:{:.4f}".format(imname_short, cval_list[i][1]))
            else:
                ann.append("")
            for c in range(10):
                try:
                    cand_img = self.datadir + row[2][c]
                    cand_cval = cval_list[i][2][c]
                    imname_short = self.shorten_imgname(row[2][c])
                    r.append(cand_img)
                    if annotate:
                        ann.append("{}:{:.4f}".format(imname_short, cand_cval))
                    else:
                        ann.append("")
                except:
                    r.append(blank_img)
                    ann.append("")
            imnames.append(r)
            annotations.append(ann)
        self.fill_panel(imnames, ylabels, annotations)
        self.figure.suptitle(
            "Axis: {}, Exclusion Distance: {}".format(ax, exc_dist), fontsize=15, y=0.9
        )


class ImagePanelAllCandidates(ImageGridV2):
    def __init__(self, imgdir: str, row_num: int):
        ImageGridV2.__init__(self, row_num, 11)
        self.imgdir = imgdir

    def shorten_imgname(self, imgname: str) -> str:
        obj_codes = imgname.split("_")
        series_name = obj_codes[1].split("-")[1]
        series_name = series_name.split(".")[0]
        return series_name

    def change_border_color(self, border_colors: list) -> None:
        for i, ax in enumerate(self.grid):
            r = i // self.col
            c = i % self.col
            ax.set_axis_on()
            matplotlib.pyplot.setp(
                ax.spines.values(), color=border_colors[r][c], linewidth=3
            )

    def fill_all_candidate_panel(
        self, refimg: str, candidates: np.ndarray, candidate_cvals: np.ndarray
    ) -> None:
        cmap = cm.get_cmap("plasma")
        candidates = np.reshape(candidates, (self.row, self.col))
        candidate_imgnames = np.char.add(self.imgdir, candidates)
        candidate_imgnames = candidate_imgnames.tolist()
        candidates = candidates.tolist()

        candidate_cvals = np.reshape(candidate_cvals, (self.row, self.col))
        candidate_cvals = candidate_cvals.tolist()
        annotations = []
        ylabels = []
        border_colors = []
        for r, row in enumerate(candidate_cvals):
            ann = []
            bc = []
            for c, cval in enumerate(row):
                shortened_imgname = self.shorten_imgname(candidates[r][c])
                ann.append("{}:{:.4f}".format(shortened_imgname, cval))
                if c == 0:
                    ylabels.append(re.findall("[a-zA-Z]+", shortened_imgname)[0])
                if candidates[r][c] == refimg:
                    bc.append((1.0, 0.0, 0.0, 1.0))
                else:
                    bc.append(cmap(cval))
            annotations.append(ann)
            border_colors.append(bc)
        self.fill_panel(candidate_imgnames, ylabels, annotations)
        self.change_border_color(border_colors)
        obj_codes = refimg.split("_")
        series_name = obj_codes[1].split("-")[1]
        shortened_imgname = (
            obj_codes[0] + "_" + obj_codes[1][0:4] + "-{}".format(series_name)
        )
        self.figure.suptitle(
            "All Candidate Images for {}".format(shortened_imgname), fontsize=15, y=0.9
        )


class ImagePanelSameObjBestMatchingWithExcDist(ImageGridV2):
    def __init__(self, imgdir: str, num_cols: int):
        ImageGridV2.__init__(self, 1, num_cols)
        self.imgdir = imgdir
        self.numsample = num_cols

    def shorten_imgname(self, imgname: str) -> str:
        obj_codes = imgname.split("_")
        series_name = obj_codes[1].split("-")[1]
        series_name = series_name.split(".")[0]
        return series_name

    def set_xlabel(self, dists: list) -> None:
        for i, ax in enumerate(self.grid):
            r = i // self.col
            c = i % self.col
            if r == 0:
                if c == 0:
                    ax.set_title("Reference Img", fontsize=10, fontweight="bold")
                else:
                    ax.set_title(
                        "Dist = {}".format(dists[c - 1]), fontsize=10, fontweight="bold"
                    )

    def fill_best_matching_panel(
        self,
        refimg: str,
        candidates_sorted: np.ndarray,
        candidate_cvals_sorted: np.ndarray,
        dists_sorted: np.ndarray,
        excdist: int,
        axis: str,
    ) -> None:
        candidate_imgnames = np.char.add(self.imgdir, candidates_sorted)
        candidate_imgnames = candidate_imgnames.tolist()
        candidates = candidates_sorted.tolist()
        dists = dists_sorted.tolist()

        candidate_cvals = candidate_cvals_sorted.tolist()
        annotations = []
        # add ref img to first
        candidate_imgnames.insert(0, self.imgdir + refimg)
        candidate_cvals.insert(0, 1.0)
        candidates.insert(0, refimg)
        ann = []
        for c, cval in enumerate(candidate_cvals):
            if c < self.numsample:
                shortened_imgname = self.shorten_imgname(candidates[c])
                ann.append("{}:{:.4f}".format(shortened_imgname, cval))
            else:
                break
        annotations.append(ann)
        ylabels = ["Sorted Matches"]
        candidate_imgnames = [candidate_imgnames]
        self.fill_panel(candidate_imgnames, ylabels, annotations)
        self.figure.suptitle(
            "Top {} matches with Exclusion Distance of {} in {}".format(
                self.numsample - 1, excdist, axis
            ),
            y=1.2,
        )
        self.set_xlabel(dists)


class ImagePanelSingleSample(ImageGridV2):
    def __init__(self, imgdir):
        ImageGridV2.__init__(self, 4, 11)
        self.imgdir = imgdir

    def set_xlabel(self):
        for i, ax in enumerate(self.grid):
            r = i // self.col
            c = i % self.col
            if r == 0:
                ax.set_title(
                    "Exc. Dist = {}".format(c),
                    fontsize=10,
                    color="red",
                    fontweight="bold",
                )
                ax.spines["bottom"].set_color("red")
                ax.spines["bottom"].set_linewidth(4)
            if r == 1:
                ax.spines["top"].set_color("blue")
                ax.spines["top"].set_linewidth(4)
            if c == 0:
                ax.spines["right"].set_color("red")
                ax.spines["right"].set_linewidth(6)

    def set_second_xlabel(self):
        for i, ax in enumerate(self.grid):
            r = i // self.col
            c = i % self.col
            if r == 3:
                if c == 0:
                    text = "Reference Image"
                else:
                    text = "N = {}".format(c)
                ax.set_xlabel(text, fontsize=10, color="blue", fontweight="bold")

    def set_ylabel_color(self):
        ylabel_colors = ["red", "blue", "black", "green"]
        for i, ax in enumerate(self.grid):
            r = i // self.col
            c = i % self.col
            if c == 0:
                ax.yaxis.label.set_color(ylabel_colors[r])

    def fill_single_sample_panel(self, sampled_imgname, row_infos):
        blank_dir = PROJECT_DIR
        imnames = []
        annotations = []
        obj_codes = sampled_imgname.split("_")
        series_name = obj_codes[1].split("-")[1]
        shortened_imgname = (
            obj_codes[0] + "_" + obj_codes[1][0:4] + "-{}".format(series_name)
        )
        for i, item in enumerate(row_infos):
            if i == 0:
                row = list(item[1])
                row = [
                    blank_dir + img if img == "blank.png" else self.imgdir + img
                    for img in row
                ]
                reference_img = row[0]
                imnames.append(row)
                ann = []
                for idx, img in enumerate(item[1]):
                    if img != "blank.png":
                        series_name = img.split("_")[1]
                        series_name = series_name.split("-")[1]
                        series_name = series_name.split(".")[0]
                        ann.append("{}:{:.4f}".format(series_name, item[0][idx]))
                    else:
                        ann.append("")
                ann[0] = shortened_imgname
                annotations.append(ann)
            else:
                row = [self.imgdir + img for img in item[1]]
                row.insert(0, reference_img)
                imnames.append(row)
                ann = []
                for idx, img in enumerate(item[1]):
                    objname = img.split("_")
                    objcode = objname[0] + "_" + objname[1].split("-")[0][0:4]
                    ann.append("{}:{:.4f}".format(objcode, item[0][idx]))
                ann.insert(0, shortened_imgname)
                annotations.append(ann)

        ylabels = [
            "Within Same Obj",
            "Top 11 Any Other Obj",
            "Top 11 Per Obj",
            "Top 11 Per Obj Category",
        ]
        self.fill_panel(imnames, ylabels, annotations)
        self.set_xlabel()
        self.set_second_xlabel()
        self.set_ylabel_color()


class ImagePanelTopPer(ImageGridV2):
    def __init__(self, imgdir):
        super().__init__(11, 12, axis_pad=0.07)
        self.imgdir = imgdir

    def shorten_imgname(self, imgname):
        obj_codes = imgname.split("_")
        series_name = obj_codes[1].split("-")[1]
        series_name = series_name.split(".")[0]
        shortened_imgname = (
            obj_codes[0] + "_" + obj_codes[1][0:4] + "-{}".format(series_name)
        )
        return shortened_imgname

    def change_border_color(self, border_colors):
        for i, ax in enumerate(self.grid):
            r = i // self.col
            c = i % self.col
            ax.set_axis_on()
            matplotlib.pyplot.setp(
                ax.spines.values(), color=border_colors[r][c], linewidth=2
            )

    def fill_top_per_panel(
        self, sampled_imgnames, row_infos, axis, exclusion_dist, obj_cat=False
    ):
        blank_dir = PROJECT_DIR
        imnames = []
        annotations = []
        border_colors = []
        for i, item in enumerate(row_infos):
            row = list(item[1])
            row.insert(0, sampled_imgnames[i])
            obj_ref = sampled_imgnames[i].split("-")[0]
            if obj_cat:
                obj_ref = obj_ref.split("_")[0]
            # red if correct obj / objcat
            border_color_row = ["red" if obj_ref in img else "white" for img in row]
            not_same_obj_idx = [i for i, n in enumerate(row) if obj_ref not in n][0]
            if row[not_same_obj_idx] != "blank.png":
                if obj_cat:
                    border_color_row[not_same_obj_idx] = "green"
                else:
                    border_color_row[not_same_obj_idx] = "blue"
            border_colors.append(border_color_row)
            row = [
                blank_dir + img if img == "blank.png" else self.imgdir + img
                for img in row
            ]
            imnames.append(row)
            ann = []
            for idx, img in enumerate(item[1]):
                if img != "blank.png":
                    shortened_imgname = self.shorten_imgname(img)
                    ann.append("{}:{:.4f}".format(shortened_imgname, item[0][idx]))
                else:
                    ann.append("")
            ann.insert(0, "")
            annotations.append(ann)

        ylabels = [self.shorten_imgname(imname) for imname in sampled_imgnames]
        self.fill_panel(imnames, ylabels, annotations)
        if not obj_cat:
            self.figure.suptitle(
                "Top Per Object with Exclusion Distance of {} in {}".format(
                    exclusion_dist, axis
                ),
                fontsize=15,
                y=0.9,
            )
        else:
            self.figure.suptitle(
                "Top Per Object Category with Exclusion Distance of {} in {}".format(
                    exclusion_dist, axis
                ),
                fontsize=15,
                y=0.9,
            )
        self.change_border_color(border_colors)



from tqdm import tqdm
import numpy as np
from itertools import combinations
import cupy as cp
from cupyx.scipy.linalg import tri
import functools
from shapey.utils.customdataset import ImageFolderWithPaths, PermutationPairsDataset
from shapey.utils.modelutils import GetModelIntermediateLayer
from shapey.utils.customfunc import pearsonr_batch, ln_batch
import torchvision.transforms as transforms
import torch
import torchvision.models as models
from torch.utils.data import Subset
from typing import Tuple
from h5py import File
import logging
import traceback

log = logging.getLogger(__name__)


def extract_features_resnet50(datadir: str) -> Tuple[list, list]:
    normalize = transforms.Normalize(
        mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]
    )

    dataset = ImageFolderWithPaths(
        datadir,
        transforms.Compose([transforms.Resize(224), transforms.ToTensor(), normalize]),
    )

    data_loader = torch.utils.data.DataLoader(
        dataset, batch_size=1, shuffle=False, num_workers=0, pin_memory=True
    )
    resnet50 = models.resnet50(pretrained=True)
    resnet50_gap = GetModelIntermediateLayer(resnet50, -1)
    resnet50_gap.cuda().eval()

    # compute features
    original_stored_imgname = []
    original_stored_feat = []
    for s in tqdm(data_loader):
        img1, _, fname1 = s
        fname1 = fname1[0].split("/")[-1]
        output1 = resnet50_gap(img1.cuda())
        output1 = torch.flatten(output1)
        output1_store = output1.cpu().data.numpy()
        original_stored_imgname.append(fname1)
        original_stored_feat.append(output1_store)
    return original_stored_imgname, original_stored_feat


def compute_correlation_and_save(
    permutation_dataset: PermutationPairsDataset,
    hdfstore: File,
    corrval_key: str,
    batch_size: int = 20000,
    num_workers: int = 8,
) -> None:
    data_loader = torch.utils.data.DataLoader(
        permutation_dataset,
        batch_size=batch_size,
        shuffle=False,
        pin_memory=True,
        num_workers=num_workers,
    )
    log.info("Computing feature correlations...")
    completed = False
    while not completed:
        try:
            for s1, s2 in tqdm(data_loader):
                idx1, feat1 = s1
                idx2, feat2 = s2
                idx1 = idx1.data.numpy()
                idx2 = idx2.data.numpy()
                # feat1 = hdfstore[feature_output_key][idx1, :]
                # feat2 = hdfstore[feature_output_key][idx2, :]
                # feat1 = torch.tensor(feat1).cuda()
                # feat2 = torch.tensor(feat2).cuda()

                # compute correlation
                rval = pearsonr_batch(feat1.cuda(), feat2.cuda())
                data = hdfstore[corrval_key][idx1.min() : idx1.max() + 1, :]
                data[idx1 - idx1.min(), idx2] = rval.cpu().data.numpy().flatten()
                hdfstore[corrval_key][idx1.min() : idx1.max() + 1, :] = data
                log.info("Last computed: idx1: {}, idx2: {}".format(idx1[-1], idx2[-1]))
            completed = True
        except Exception as e:
            log.error(e)
            log.error(traceback.format_exc())
            log.info(
                "Last batch computed: ({}, {}) ~ ({}, {})".format(
                    idx1[0], idx2[0], idx1[-1], idx2[-1]
                )
            )
        finally:
            del data_loader
            if not completed:
                log.info(
                    "Restarting data loader from ({}, {})...".format(idx1[0], idx2[0])
                )
                idx = idx1[0] * permutation_dataset.datalen + idx2[0]
                new_dataset = Subset(
                    permutation_dataset, range(idx, len(permutation_dataset))
                )
                data_loader = torch.utils.data.DataLoader(
                    new_dataset,
                    batch_size=batch_size,
                    shuffle=False,
                    pin_memory=True,
                    num_workers=num_workers,
                )


def compute_distance_and_save(
    permutation_dataset: PermutationPairsDataset,
    hdfstore: File,
    corrval_key: str,
    batch_size: int = 20000,
    num_workers: int = 8,
    distance: str = "correlation",
) -> None:
    data_loader = torch.utils.data.DataLoader(
        permutation_dataset,
        batch_size=batch_size,
        shuffle=False,
        pin_memory=True,
        num_workers=num_workers,
    )

    log.info("Computing distances using {}...".format(distance))
    completed = False
    error_idx = (0, 0, 0, 0)
    repeat_count = 0
    while not completed:
        try:
            for s1, s2 in tqdm(data_loader):
                idx1, feat1 = s1
                idx2, feat2 = s2
                idx1 = idx1.data.numpy()
                idx2 = idx2.data.numpy()
                # feat1 = hdfstore[feature_output_key][idx1, :]
                # feat2 = hdfstore[feature_output_key][idx2, :]
                # feat1 = torch.tensor(feat1).cuda()
                # feat2 = torch.tensor(feat2).cuda()

                # compute correlation
                if distance == "correlation":
                    rval = pearsonr_batch(feat1.cuda(), feat2.cuda())
                elif distance == "l1":
                    rval = ln_batch(feat1.cuda(), feat2.cuda(), n=1)
                elif distance == "l2":
                    rval = ln_batch(feat1.cuda(), feat2.cuda(), n=2)
                data = hdfstore[corrval_key][idx1.min() : idx1.max() + 1, :]
                data[idx1 - idx1.min(), idx2] = rval.cpu().data.numpy().flatten()
                hdfstore[corrval_key][idx1.min() : idx1.max() + 1, :] = data
                log.info("Last computed: idx1: {}, idx2: {}".format(idx1[-1], idx2[-1]))
            completed = True
        except Exception as e:
            log.error(e)
            log.error(traceback.format_exc())
            log.info(
                "Last batch computed: ({}, {}) ~ ({}, {})".format(
                    idx1[0], idx2[0], idx1[-1], idx2[-1]
                )
            )
            if error_idx == (idx1[0], idx2[0], idx1[-1], idx2[-1]):
                log.info("repeating the same batch...")
                repeat_count += 1

        finally:
            del data_loader
            if not completed:
                if repeat_count > 1000:
                    log.error("repeating the same batch over and over...")
                    log.info("exiting...")
                    break
                log.info(
                    "Restarting data loader from ({}, {})...".format(idx1[0], idx2[0])
                )
                error_idx = (idx1[0], idx2[0], idx1[-1], idx2[-1])
                idx = idx1[0] * permutation_dataset.datalen + idx2[0]
                new_dataset = Subset(
                    permutation_dataset, range(idx, len(permutation_dataset))
                )
                data_loader = torch.utils.data.DataLoader(
                    new_dataset,
                    batch_size=batch_size,
                    shuffle=False,
                    pin_memory=True,
                    num_workers=num_workers,
                )


class ImgCorrelationDataProcessorV2:
    def __init__(self, hdfstore: File) -> None:
        self.imgnames = hdfstore["/feature_output/imgname"][:].astype("U")
        self.axes_of_interest = self.generate_axes_of_interest()
        self.objnames = np.unique(np.array([c.split("-")[0] for c in self.imgnames]))

    @staticmethod
    def generate_axes_of_interest() -> list:
        axes = ["x", "y", "p", "r", "w"]
        axis_of_interest = []
        for choose in range(1, 7):
            for comb in combinations(axes, choose):
                axis_of_interest.append(functools.reduce(lambda a, b: a + b, comb))
        axis_of_interest.sort()
        return axis_of_interest

    def cut_corrmat(
        self,
        hdfstore: File,
        num_objs: int,
        post_processed: bool = False,
        pp_exclusion: str = "soft",
    ) -> None:
        if not hasattr(self, "objnames_buffer"):
            self.objnames_buffer = self.objnames
            self.imgnames_buffer = self.imgnames
        obj_per_category = int(num_objs / 20)
        obj_idxs = np.array(
            [
                v + 10 * cat
                for cat in range(20)
                for v in np.sort(np.random.choice(10, obj_per_category, replace=False))
            ]
        )
        self.objnames = self.objnames_buffer[obj_idxs]

        # cut image names
        img_idxs_start = obj_idxs * 11 * 31
        img_idxs_end = (obj_idxs + 1) * 11 * 31
        img_idxs = np.array(
            [np.array([range(a, b)]) for (a, b) in zip(img_idxs_start, img_idxs_end)]
        ).flatten()
        self.imgnames = self.imgnames_buffer[img_idxs]

        # cut corr matrix
        if post_processed:
            cval_matrix = hdfstore["/pairwise_correlation/postprocessed"]
            key_head = "/pairwise_correlation/postprocessed/{}".format(pp_exclusion)
            if pp_exclusion == "hard":
                cval_orig = hdfstore["/pairwise_correlation/original"]
        else:
            cval_matrix = hdfstore["/pairwise_correlation/original"]
            key_head = "/pairwise_correlation/original"

        mat_list = []
        for i in obj_idxs:
            obj_mat_list = []
            for j in obj_idxs:
                obj_mat = cval_matrix[
                    i * 11 * 31 : (i + 1) * 11 * 31, j * 11 * 31 : (j + 1) * 11 * 31
                ]
                obj_mat_list.append(obj_mat)
            obj_mat_row = np.concatenate(obj_mat_list, axis=1)
            mat_list.append(obj_mat_row)
        mat = np.concatenate(mat_list)
        try:
            hdfstore[key_head + "_{}".format(num_objs)] = mat
            hdfstore[key_head + "_{}_obj_idx".format(num_objs)] = obj_idxs
        except RuntimeError:
            pass

    def excluded_to_zero(
        self, corr_mat_sameobj: cp.ndarray, axis: str, exc_dist: int, pure: bool = True, distance: str = "correlation"
    ) -> cp.ndarray:
        # corr_mat_obj has to be a cut-out copy of the original matrix!!!
        # create list with axis of interest in the alphabetical order

        if exc_dist != 0:
            # first create a 11x11 sampling mask per axis
            sampling_mask = 1 - (
                tri(11, 11, exc_dist - 1, dtype=float)
                - tri(11, 11, -exc_dist, dtype=float)
            )
            sampling_mask[sampling_mask == 0] = cp.nan
        # pure transform
        if pure:
            # cut out the axis of interest
            idx = self.axes_of_interest.index(axis)
            corr_mat_axis = cp.copy(
                corr_mat_sameobj[idx * 11 : (idx + 1) * 11, idx * 11 : (idx + 1) * 11]
            )
            # sample with the sampling mask
            if exc_dist != 0:
                corr_mat_axis = cp.multiply(corr_mat_axis, sampling_mask)
            return corr_mat_axis
        elif exc_dist != 0:
            contain_ax = cp.array(
                [
                    [
                        cp.array([c in a for c in axis]).all()
                        for a in self.axes_of_interest
                    ]
                ],
                dtype=int,
            )
            # selects relevant axis
            repeat_mask = contain_ax * contain_ax.T
            # create sampling mask of size 11 (# image in each series) x 31 (total number of axes)
            repeat_mask = cp.repeat(cp.repeat(repeat_mask, 11, axis=1), 11, axis=0)
            sampling_mask_whole = cp.tile(sampling_mask, (31, 31))
            sampling_mask_whole = cp.multiply(sampling_mask_whole, repeat_mask)
            if distance != "correlation":
                sampling_mask_whole[sampling_mask_whole==0] = cp.nan
            # sample from the correlation matrix using the sampling mask
            corr_mat_sameobj = cp.multiply(sampling_mask_whole, corr_mat_sameobj)
            # cut out only the samples where starting image is contained within the transformation series "axis"
            idx = self.axes_of_interest.index(axis)
            corr_mat_sameobj = corr_mat_sameobj[idx * 11 : (idx + 1) * 11, :]
            return corr_mat_sameobj
        else:
            idx = self.axes_of_interest.index(axis)
            corr_mat_sameobj = corr_mat_sameobj[idx * 11 : (idx + 1) * 11, :]
            return corr_mat_sameobj

    def get_coord_corrmat(self, obj_name: str, ax: str = "all") -> Tuple[int, int]:
        idx = np.where(self.objnames == obj_name)[0][0]

        if ax == "all":
            return idx * 11 * 31, (idx + 1) * 11 * 31
        else:
            ax_idx = self.axes_of_interest.index(ax)
            return idx * 11 * 31 + ax_idx * 11, idx * 11 * 31 + (ax_idx + 1) * 11

    def get_top1_cval_other_object(
        self,
        cval_matrix_hdf: File,
        obj: str,
        ax: str,
        cval_arr_sameobj: np.ndarray,
        distance="correlation",
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        c1, c2 = self.get_coord_corrmat(obj, ax=ax)
        cval_mat_np = cp.array(cval_matrix_hdf[c1:c2, :])
        # zero masking the same object
        c3, c4 = self.get_coord_corrmat(obj)
        if distance == "correlation":
            cval_mat_np[:, c3:c4] = 0
        else:
            cval_mat_np[:, c3:c4] = np.nan

        if distance == "correlation":
            # get top1 cval per row
            top1_cval = cval_mat_np.max(axis=1)
            top1_idx = cval_mat_np.argmax(axis=1)
            # get image rank
            sameobj_imagerank = []
            for col in cval_arr_sameobj.T:
                comparison_mask = cp.tile(col, (cval_mat_np.shape[1], 1)).T
                count_col = (cval_mat_np > comparison_mask).sum(axis=1)
                count_col = count_col.get()
                count_col = count_col.astype(np.float)
                count_col[np.isnan(col)] = np.nan
                sameobj_imagerank.append(count_col)
        else:
            # get top1 closest per row
            top1_cval = cp.nanmin(cval_mat_np, axis=1)
            top1_idx = cp.nanargmin(cval_mat_np, axis=1)
            # get image rank
            sameobj_imagerank = []
            for col in cval_arr_sameobj.T:
                comparison_mask = cp.tile(col, (cval_mat_np.shape[1], 1)).T
                count_col = (cval_mat_np < comparison_mask).sum(axis=1)
                count_col = count_col.get()
                count_col = count_col.astype(np.float)
                count_col[np.isnan(col)] = np.nan
                sameobj_imagerank.append(count_col)
        return top1_idx.get(), top1_cval.get(), np.array(sameobj_imagerank).T

    def get_top_per_object(
        self, cval_matrix_hdf: File, obj: str, ax: str, distance: str = "correlation"
    ) -> Tuple[np.ndarray, np.ndarray]:
        c1, c2 = self.get_coord_corrmat(obj, ax=ax)
        cval_mat_np = cval_matrix_hdf[c1:c2, :]
        top1_cvals = []
        top1_idxs = []
        for o in self.objnames:
            if not o == obj:
                c3, c4 = self.get_coord_corrmat(o)
                cval_mat_obj = cval_mat_np[:, c3:c4]
                if distance == "correlation":
                    top1_cvals.append(cval_mat_obj.max(axis=1))
                    top1_idxs.append(cval_mat_obj.argmax(axis=1))
                else:
                    top1_cvals.append(cval_mat_obj.min(axis=1))
                    top1_idxs.append(cval_mat_obj.argmin(axis=1))
        return (
            np.array(top1_idxs, dtype=np.int64).T,
            np.array(top1_cvals, dtype=float).T,
        )

    def get_top1_sameobj_with_exclusion(
        self, obj, ax, cval_matrix, pure=False, distance: str = "correlation"
    ):
        c1, c2 = self.get_coord_corrmat(obj)
        cval_sameobj = cp.asarray(cval_matrix[c1:c2, c1:c2])
        max_cvals = []
        max_idxs = []
        for xdist in range(0, 11):
            res = self.excluded_to_zero(cval_sameobj, ax, xdist, pure=pure, distance=distance)
            if distance == "correlation":
                max_cvals.append(cp.nanmax(res, axis=1))
                max_idxs.append(cp.nanargmax(res, axis=1))
            else:
                max_cvals.append(cp.nanmin(res, axis=1))
                max_idxs.append(cp.nanargmin(res, axis=1))
        max_cvals = cp.array(max_cvals, dtype=float).T
        max_idxs = cp.array(max_idxs, dtype=cp.int64).T
        return max_cvals.get(), max_idxs.get()

    def get_objrank(
        self,
        cval_arr_sameobj: np.ndarray,
        top1_cval_per_obj: np.ndarray,
        distance: str = "correlation",
    ) -> np.ndarray:
        sameobj_objrank = []
        top1_cval_per_obj = cp.array(top1_cval_per_obj)
        for col in cval_arr_sameobj.T:
            comparison_mask = cp.tile(col, (top1_cval_per_obj.shape[1], 1)).T
            if distance == "correlation":
                count_col = (top1_cval_per_obj > comparison_mask).sum(axis=1)
            else:
                count_col = (top1_cval_per_obj < comparison_mask).sum(axis=1)
            count_col = count_col.get()
            count_col = count_col.astype(np.float)
            count_col[np.isnan(col)] = np.nan
            sameobj_objrank.append(count_col)
        return np.array(sameobj_objrank).T

    def get_top1_objcat_with_exclusion(
        self,
        obj_ref,
        obj_comp,
        ax,
        cval_matrix,
        pure=False,
        distance: str = "correlation",
    ):
        c1, c2 = self.get_coord_corrmat(obj_ref)
        c3, c4 = self.get_coord_corrmat(obj_comp)
        cval_sameobjcat = cp.asarray(cval_matrix[c1:c2, c3:c4])
        max_cvals = []
        max_idxs = []
        for xdist in range(0, 11):
            res = self.excluded_to_zero(cval_sameobjcat, ax, xdist, pure=pure)
            if distance == "correlation":
                max_cvals.append(cp.nanmax(res, axis=1))
                max_idxs.append(cp.nanargmax(res, axis=1))
            else:
                max_cvals.append(cp.nanmin(res, axis=1))
                max_idxs.append(cp.nanargmin(res, axis=1))
        return np.array(max_cvals, dtype=float).T, np.array(max_idxs, dtype=np.int64).T

    def exclusion_distance_analysis(
        self,
        hdfstore: File,
        contrast_reversed: bool = False,
        exclusion_mode: str = "soft",
        pure: bool = False,
        num_objs: int = 0,
        distance: str = "correlation",
    ) -> None:
        if num_objs == 0:
            if contrast_reversed:
                cval_matrix = hdfstore["/pairwise_correlation/contrast_reversed"]
                key_head = "contrast_reversed/{}".format(exclusion_mode)
                if exclusion_mode == "hard":
                    cval_orig = hdfstore["/pairwise_correlation/original"]
            else:
                cval_matrix = hdfstore["/pairwise_correlation/original"]
                key_head = "original"
        else:
            if contrast_reversed:
                cval_matrix = hdfstore[
                    "/pairwise_correlation/contrast_reversed_{}".format(num_objs)
                ]
                key_head = "contrast_reversed_{}/{}".format(num_objs, exclusion_mode)
                if exclusion_mode == "hard":
                    cval_orig = hdfstore[
                        "/pairwise_correlation/original_{}".format(num_objs)
                    ]
            else:
                cval_matrix = hdfstore[
                    "/pairwise_correlation/original_{}".format(num_objs)
                ]
                key_head = "original_{}".format(num_objs)

        for obj in tqdm(self.objnames):
            obj_cat = obj.split("_")[0]
            for ax in self.axes_of_interest:
                obj_ax_key = "/" + key_head + "/" + obj + "/" + ax
                try:
                    hdfstore.create_group(obj_ax_key)
                except ValueError:
                    log.info(obj_ax_key + " already exists")
                # make same object cval array with exclusion distance in ax
                cval_arr_sameobj, idx_sameobj = self.get_top1_sameobj_with_exclusion(
                    obj, ax, cval_matrix, pure=pure, distance=distance
                )
                hdfstore[obj_ax_key + "/top1_cvals"] = cval_arr_sameobj
                hdfstore[obj_ax_key + "/top1_idx"] = idx_sameobj
                if not contrast_reversed:
                    cval_mat_name = "cval_matrix"
                else:
                    if exclusion_mode == "soft":
                        cval_mat_name = "cval_matrix"
                    elif exclusion_mode == "hard":
                        cval_mat_name = "cval_orig"
                # grab top1 for all other objects
                (
                    top1_idx_otherobj,
                    top1_cval_otherobj,
                    sameobj_imagerank,
                ) = self.get_top1_cval_other_object(
                    locals()[cval_mat_name],
                    obj,
                    ax,
                    cval_arr_sameobj,
                    distance=distance,
                )
                hdfstore[obj_ax_key + "/top1_cvals_otherobj"] = top1_cval_otherobj
                hdfstore[obj_ax_key + "/top1_idx_otherobj"] = top1_idx_otherobj
                # count how many images come before the top1 same object view with exclusion
                hdfstore[obj_ax_key + "/sameobj_imgrank"] = sameobj_imagerank

                # grab top per object
                top1_per_obj_idxs, top1_per_obj_cvals = self.get_top_per_object(
                    locals()[cval_mat_name], obj, ax, distance=distance
                )
                hdfstore[obj_ax_key + "/top1_per_obj_cvals"] = top1_per_obj_cvals
                hdfstore[obj_ax_key + "/top1_per_obj_idxs"] = top1_per_obj_idxs
                # count how many objects come before the same object view with exclusion
                sameobj_objrank = self.get_objrank(
                    cval_arr_sameobj, top1_per_obj_cvals, distance=distance
                )
                hdfstore[obj_ax_key + "/sameobj_objrank"] = sameobj_objrank

                # for object category exclusion analysis
                same_obj_cat_key = obj_ax_key + "/same_cat"
                for o in self.objnames:
                    other_obj_cat = o.split("_")[0]
                    if other_obj_cat == obj_cat and o != obj:
                        (
                            cval_arr_sameobjcat,
                            idx_sameobjcat,
                        ) = self.get_top1_objcat_with_exclusion(
                            obj, o, ax, cval_matrix, pure=pure, distance=distance
                        )
                        hdfstore[
                            same_obj_cat_key + "/{}/top1_cvals".format(o)
                        ] = cval_arr_sameobjcat
                        hdfstore[
                            same_obj_cat_key + "/{}/top1_idx".format(o)
                        ] = idx_sameobjcat



import setuptools

# reading long description from file
with open("README.md") as file:
    long_description = file.read()


# specify requirements of your package here
REQUIREMENTS = [
    "torch >= 1.10.0",
    "torchvision >= 0.11.2",
    "numpy>=1.18.4",
    "kaleido>=0.0.1",
    "tqdm>=4.48.0",
    "cupy-cuda102",
    "matplotlib>=3.3.1",
    "gdown>=4.0.2",
    "h5py>=2.10.0",
]

# some more details
CLASSIFIERS = [
    "License :: OSI Approved :: MIT License",
    "Programming Language :: Python :: 3.6",
    "Development Status :: 4 - Beta",
]

# calling the setup function
setuptools.setup(
    name="ShapeY",
    version="0.1.5",
    description="Benchmark that tests shape recognition",
    long_description=long_description,
    url="https://github.com/njw0709/ShapeY",
    author="Jong Woo Nam",
    author_email="namj@usc.edu",
    license="MIT",
    packages=setuptools.find_packages(),
    classifiers=CLASSIFIERS,
    install_requires=REQUIREMENTS,
    keywords="tests shape recognition capacity",
)



import subprocess
import functools
from itertools import combinations
import numpy as np


def run_script(command, **kwargs):
    for k in kwargs.keys():
        command = command + " " + "--" + str(k) + " " + str(kwargs[k])
    command = command.split(" ")
    print(command)
    subprocess.run(command)
    print("Done with " + str(command))


def check_image_order(
    user_img_list: np.ndarray, reference_img_list: np.ndarray
) -> bool:
    return np.all(user_img_list == reference_img_list)


def make_axis_of_interest() -> list:
    axes = ["x", "y", "p", "r", "w"]
    axis_of_interest = []
    for choose in range(1, 4):
        for comb in combinations(axes, choose):
            axis_of_interest.append(functools.reduce(lambda a, b: a + b, comb))
    axis_of_interest.sort()
    return axis_of_interest



import matplotlib.pyplot as plt
import numpy as np


def make_corr_fall_off_single_sample(row_info, excaxis):
    figure, ax = plt.subplots(figsize=(1.6 * 5, 1.6 * 4))

    best_matching_x = np.linspace(1, 11, 11)
    ax.set_xticks(best_matching_x)
    ax.hlines(row_info[1][0][0], 0.5, 11.5, colors="b", linestyle="--", linewidth=2)
    ax.hlines(row_info[3][0][0], 0.5, 11.5, colors="g", linestyle="--", linewidth=2)
    ax.plot(
        best_matching_x,
        row_info[1][0],
        "bv-",
        label="Best Matching Any Obj Except Original",
    )
    ax.plot(
        best_matching_x,
        row_info[2][0],
        "k^-",
        label="Best Matching Per Obj Except Original",
    )
    ax.plot(
        best_matching_x,
        row_info[3][0],
        "gx-",
        label="Best Matching Per Obj Category Except Original Category",
    )
    ax.set_xlabel("Nth Best Matching", color="blue", fontweight="bold")
    ax.tick_params(axis="x", labelcolor="blue")
    ax.set_xlim(0.5, 11.5)

    ax2 = ax.twiny()
    ax2.plot(
        row_info[0][0],
        "r.-",
        label="Correlation Fall Off with Exclusion in {}".format(excaxis),
        markersize=10,
    )

    exclusion_x = np.linspace(0, 10, 11)
    yticks = np.linspace(0.2, 1.0, 9)
    ax2.set_xticks(exclusion_x)
    ax2.set_yticks(yticks)
    ax2.set_xlabel(
        "Exclusion Distance in {}".format(excaxis), color="red", fontweight="bold"
    )
    ax2.tick_params(axis="x", labelcolor="red")
    ax2.set_ylabel("Correlation Values")
    ax2.set_xlim(-0.5, 10.5)
    ax2.set_ylim(0.2, 1.1)
    ax.grid()

    ax2.spines["top"].set_color("red")
    ax2.spines["top"].set_linewidth(2)
    ax2.spines["bottom"].set_color("blue")
    ax2.spines["bottom"].set_linewidth(2)

    figure.legend(bbox_to_anchor=(0.32, 0.28), loc="upper left", ncol=1)
    return figure


def draw_correlation_fall_off_with_exclusion(
    axes, objname, top_per_obj_row_infos, objcat=False
):
    x = np.linspace(1, 10, 10)
    if objcat:
        wrong_color = "green"
    else:
        wrong_color = "blue"
    error_count = 0
    total_count = 0
    for (cvals, imnames) in top_per_obj_row_infos:
        if imnames[0] != "blank.png":
            total_count += 1
            cvals = cvals[:10]
            axes.plot(x, cvals, "-", c="gray", alpha=0.8, linewidth=0.7, zorder=1)
            issameobj = np.array([objname in img for img in imnames])
            if not issameobj[0]:
                error_count += 1
            axes.scatter(
                x[~issameobj[:10]],
                cvals[~issameobj[:10]],
                edgecolor="gray",
                facecolors="gray",
                s=10,
                zorder=1,
            )
            axes.scatter(
                x[~issameobj[:10]][0],
                cvals[~issameobj[:10]][0],
                edgecolor=wrong_color,
                facecolors="none",
                s=10,
                zorder=2,
            )
            # axes.scatter(x[issameobj[:10]], cvals[issameobj[:10]], c='r', s=15, zorder=-1)
    for (cvals, imnames) in top_per_obj_row_infos:
        if imnames[0] != "blank.png":
            cvals = cvals[:10]
            issameobj = np.array([objname in img for img in imnames])
            axes.scatter(
                x[issameobj[:10]], cvals[issameobj[:10]], c="r", s=15, zorder=3
            )

    axes.set_xlim([0, 11])
    axes.set_ylim([0.4, 1.1])
    axes.set_xticks(x)
    error_rate = error_count / total_count
    if error_rate > 0.5:
        if objcat:
            error_color = (1.0 - (error_rate - 0.5) * 2, 1.0, 0.0)
        else:
            error_color = (1.0 - (error_rate - 0.5) * 2, 0.0, 1.0)
    else:
        if objcat:
            error_color = (1.0, error_rate * 2, 0.0)
        else:
            error_color = (1.0, 0.0, error_rate * 2)
    axes.text(
        2.5,
        1.02,
        "Error Rate: {}/{}".format(error_count, total_count),
        fontsize=12,
        color=error_color,
        fontweight="bold",
    )
    plt.setp(axes.spines.values(), color=error_color, linewidth=3)


def draw_correlation_fall_off_whole_panel(
    corr_fall_off_row_infos, current_objname, axis, exclusion_dist, objcat=False
):
    fig, axes = plt.subplots(10, 11, figsize=(3.5 * 11, 3.5 * 10))
    for r, ax_row in enumerate(axes):
        for c, ax in enumerate(ax_row):
            (sampled_imgs, top_per_obj_row_infos) = corr_fall_off_row_infos[r][c]
            objname = sampled_imgs[0].split("-")[0]
            if c == 0:
                obj_codes = objname.split("_")
                shortened_imgname = obj_codes[0] + "_" + obj_codes[1][0:4]
                ax.set_ylabel(shortened_imgname)
            if r == 0:
                ax.set_title("Exc. Dist={}".format(c))
            draw_correlation_fall_off_with_exclusion(
                ax, objname, top_per_obj_row_infos, objcat=objcat
            )
            if objname == current_objname and c == exclusion_dist:
                plt.setp(ax.spines.values(), linewidth=8)
    if objcat:
        fig.suptitle(
            "Top Per Obj Category Corr Fall Off, Object category = {}, Exclusion axis = {}".format(
                obj_codes[0], axis
            ),
            fontsize=30,
        )
    else:
        fig.suptitle(
            "Top Per Obj Corr Fall Off, Object category = {}, Exclusion axis = {}".format(
                obj_codes[0], axis
            ),
            fontsize=30,
        )
    return fig, axes
